import math
from dataclasses import MISSING, is_dataclass
from datetime import date, datetime
from enum import Enum
from itertools import pairwise
from pathlib import Path
from types import NoneType, UnionType, WrapperDescriptorType
from typing import Annotated, Any, get_args, get_origin

from ovld import (
    Code,
    CodegenInProgress,
    CodegenParameter,
    Def,
    Lambda,
    Medley,
    call_next,
    code_generator,
    ovld,
    recurse,
)
from ovld.codegen import Function
from ovld.medley import KeepLast, use_combiner
from ovld.types import All, Exactly
from ovld.utils import subtler_type

from . import formats
from .auto import Auto
from .ctx import Context, Sourced, WorkingDirectory, empty
from .exc import SchemaError, ValidationError
from .instructions import pushdown
from .model import FieldModelizable, ListModelizable, Modelizable, StringModelizable, model
from .priority import LO4, LO5, LOW, MAX, MIN, STD, STD2, STD3
from .schema import AnnotatedSchema, Schema
from .tell import tells as get_tells
from .utils import (
    JSON,
    Indirect,
    IsLiteral,
    TypeAliasType,
    UnionAlias,
    basic_type,
    clsstring,
)


class BaseImplementation(Medley):
    validate_serialize: CodegenParameter[bool] = True
    validate_deserialize: CodegenParameter[bool] = True

    def __post_init__(self):
        self._schema_cache = {}

    #######################
    # User-facing methods #
    #######################

    @use_combiner(KeepLast)
    def load(self, t, obj, ctx=empty):
        return self.deserialize(t, obj, ctx)

    @use_combiner(KeepLast)
    def dump(self, t, obj, ctx=empty, *, dest=None, format=None):
        if dest:
            dest = Path(dest)
            ctx = ctx + Sourced(origin=dest)
        serialized = self.serialize(t, obj, ctx)
        if dest:
            formats.dump(p=dest, data=serialized, suffix=format)
        elif format:
            return formats.dumps(data=serialized, suffix=format)
        else:
            return serialized

    def get_serializer(self, t, ctx=empty):
        func = self.serialize.resolve(type[t], get_origin(t) or t, type(ctx))
        return lambda obj: func(self, t, obj, ctx)

    def get_deserializer(self, t, ctx=empty):
        if issubclass(t, FieldModelizable):
            func = self.deserialize.resolve(type[t], dict, type(ctx))
        elif issubclass(t, ListModelizable):
            func = self.deserialize.resolve(type[t], list, type(ctx))
        elif issubclass(t, StringModelizable):
            func = self.deserialize.resolve(type[t], str, type(ctx))
        else:
            func = type(self).deserialize
        return lambda obj: func(self, t, obj, ctx)

    ##################
    # Global helpers #
    ##################

    @classmethod
    def subcode(
        cls, method_name, t, accessor, ctx_t, ctx_expr=Code("$ctx"), after=None, validate=None
    ):
        if isinstance(accessor, str):
            acc1 = acc2 = Code(accessor)
        else:
            acc1 = Code("__TMP := $accessor", accessor=accessor)
            acc2 = Code("__TMP")

        if validate is None:
            validate = getattr(cls, f"validate_{method_name}")
        method = getattr(cls, method_name)
        if ec := getattr(cls, f"{method_name}_embed_condition")(t):
            ec = All[ec]
        if ec is not None:
            try:
                fn = method.resolve(type[t], ec, ctx_t, after=after)
                cg = getattr(fn, "__codegen__", None)
                if cg:
                    body = cg.create_expression([None, t, accessor, ctx_expr])
                    ot = get_origin(t) or t
                    if not validate:
                        return Code(body)
                    else:
                        return Code(
                            "$body if type($acc1) is $t else $recurse($self, $t, $acc2, $ctx_expr)",
                            body=Code(body),
                            acc1=acc1,
                            acc2=acc2,
                            t=ot,
                            recurse=method,
                            ctx_expr=ctx_expr,
                        )
            except (CodegenInProgress, ValueError):  # pragma: no cover
                # This is important if we are going to inline recursively
                # a type that refers to itself down the line.
                # We currently never do that.
                pass
        return Code(
            "$method_map[$tt, type($acc1), $ctxt]($self, $t, $acc2, $ctx_expr)",
            tt=subtler_type(t),
            ctxt=ctx_t,
            t=t,
            acc1=acc1,
            acc2=acc2,
            method_map=method.map,
            ctx_expr=ctx_expr,
        )

    ########################################
    # serialize:  helpers and entry points #
    ########################################

    @classmethod
    def serialize_embed_condition(cls, t):
        if t in (int, str, bool, float, NoneType):
            return t

    @ovld(priority=MIN)
    def serialize(self, t: Any, obj: Any, ctx: Context, /):
        raise ValidationError(
            f"Cannot serialize object of type '{clsstring(type(obj))}'"
            + (f" into expected type '{clsstring(t)}'." if t is not type(obj) else ""),
            ctx=ctx,
        )

    @ovld(priority=LO5)
    def serialize(self, t: Indirect | TypeAliasType, obj: Any, ctx: Context, /):
        return recurse(t.__value__, obj, ctx)

    @ovld(priority=LOW)
    def serialize(self, t: type[Annotated], obj: Any, ctx: Context, /):
        return recurse(pushdown(t), obj, ctx)

    def serialize(self, obj: Any, /):
        return recurse(type(obj), obj, empty)

    def serialize(self, t: Any, obj: Any, /):
        return recurse(t, obj, empty)

    #########################################
    # deserialize: helpers and entry points #
    #########################################

    deserialize_embed_condition = serialize_embed_condition

    @ovld(priority=MIN)
    def deserialize(self, t: Any, obj: Any, ctx: Context, /):
        try:
            # Pass through if the object happens to already be the right type
            if t is Any or isinstance(obj, t):
                return obj
        except TypeError:  # pragma: no cover
            pass
        disp = str(obj)
        if len(disp) > (n := 30):  # pragma: no cover
            disp = disp[:n] + "..."
        if isinstance(obj, str):
            descr = f"string '{disp}'"
        else:
            descr = f"object `{disp}` of type '{clsstring(type(obj))}'"
        raise ValidationError(
            f"Cannot deserialize {descr} into expected type '{clsstring(t)}'.",
            ctx=ctx,
        )

    @ovld(priority=LO5)
    def deserialize(self, t: Indirect | TypeAliasType, obj: Any, ctx: Context, /):
        return recurse(t.__value__, obj, ctx)

    @ovld(priority=LOW)
    def deserialize(self, t: type[Annotated], obj: Any, ctx: Context, /):
        return recurse(pushdown(t), obj, ctx)

    def deserialize(self, t: Any, obj: Any, /):
        return recurse(t, obj, empty)

    ####################################
    # schema: helpers and entry points #
    ####################################

    @ovld(priority=MAX)
    def schema(self, t: Any, ctx: Context, /):
        if t not in self._schema_cache:
            self._schema_cache[t] = holder = Schema(t)
            try:
                result = call_next(t, ctx)
            except Exception:
                del self._schema_cache[t]
                raise
            holder.update(result)
        return self._schema_cache[t]

    @ovld(priority=LO5)
    def schema(self, t: Indirect | TypeAliasType, ctx: Context, /):  # pragma: no cover
        return recurse(t.__value__, ctx)

    @ovld(priority=LOW)  # pragma: no cover
    def schema(self, t: type[Annotated], ctx: Context, /):
        return recurse(pushdown(t), ctx)

    def schema(self, t: Any, /):
        return recurse(t, empty)

    ################################
    # Implementations: basic types #
    ################################

    # str

    @code_generator(priority=STD)
    def serialize(cls, t: type[str], obj: str, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[str], obj: str, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[str], ctx: Context, /):
        return {"type": "string"}

    # bool

    @code_generator(priority=STD)
    def serialize(cls, t: type[bool], obj: bool, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[bool], obj: bool, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[bool], ctx: Context, /):
        return {"type": "boolean"}

    # int

    @code_generator(priority=STD)
    def serialize(cls, t: type[int], obj: int, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[int], obj: int, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[int], ctx: Context, /):
        return {"type": "integer"}

    # float

    @code_generator(priority=STD)
    def serialize(cls, t: type[float], obj: float, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[float], obj: float, ctx: Context, /):
        return Lambda(Code("$obj"))

    # float

    @code_generator(priority=STD)
    def serialize(cls, t: type[float], obj: int, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[float], obj: int, ctx: Context, /):
        return Lambda(Code("float($obj)"))

    @ovld(priority=STD)
    def schema(self, t: type[float], ctx: Context, /):
        return {"type": "number"}

    # None

    @code_generator(priority=STD)
    def serialize(cls, t: type[NoneType], obj: NoneType, ctx: Context, /):
        return Lambda(Code("$obj"))

    @code_generator(priority=STD)
    def deserialize(cls, t: type[NoneType], obj: NoneType, ctx: Context, /):
        return Lambda(Code("$obj"))

    @ovld(priority=STD)
    def schema(self, t: type[NoneType], ctx: Context, /):
        return {"type": "null"}

    ##########################
    # Implementations: dicts #
    ##########################

    @classmethod
    def __generic_codegen_dict(cls, method, t: type[dict], obj: dict, ctx: Context, /):
        (t,) = get_args(t)
        builder = dict if method == "serialize" else get_origin(t) or t
        kt, vt = get_args(t) or (object, object)
        ctx_expr = (
            Code("$ctx.follow($objt, $obj, K)", objt=t) if hasattr(ctx, "follow") else Code("$ctx")
        )
        code = "{$kbody: $vbody for K, V in $obj.items()}"
        if builder is not dict:
            code = f"$builder({code})"
        return Lambda(
            code,
            kbody=cls.subcode(method, kt, "K", ctx, ctx_expr=ctx_expr),
            vbody=cls.subcode(method, vt, "V", ctx, ctx_expr=ctx_expr),
            builder=builder,
        )

    @code_generator(priority=STD)
    def serialize(cls, t: type[dict], obj: dict, ctx: Context, /):
        return cls.__generic_codegen_dict("serialize", t, obj, ctx)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[dict], obj: dict, ctx: Context, /):
        return cls.__generic_codegen_dict("deserialize", t, obj, ctx)

    def schema(self, t: type[dict], ctx: Context, /):
        kt, vt = get_args(t)
        if kt is not str:
            raise SchemaError(
                f"Cannot create a schema for dicts with non-string keys (found key type: `{kt}`)"
            )
        follow = hasattr(ctx, "follow")
        fctx = ctx.follow(t, None, "*") if follow else ctx
        return {"type": "object", "additionalProperties": recurse(vt, fctx)}

    #####################################
    # Implementations: FieldModelizable #
    #####################################

    @code_generator(priority=STD)
    def serialize(cls, t: type[FieldModelizable], obj: Any, ctx: Context, /):
        (orig_t,) = get_args(t)
        t = model(orig_t)
        if not t.accepts(obj):
            return None
        stmts = []
        follow = hasattr(ctx, "follow")
        for f in t.fields:
            if f.property_name is None:
                raise SchemaError(
                    f"Cannot serialize '{clsstring(t)}' because its model does not specify how to serialize property '{f.name}'"
                )
            ctx_expr = (
                Code("$ctx.follow($objt, $obj, $fld)", objt=orig_t, fld=f.name)
                if follow
                else Code("$ctx")
            )
            stmt = Code(
                f"v_{f.name} = $setter",
                setter=cls.subcode(
                    "serialize", f.type, Code(f"$obj.{f.property_name}"), ctx, ctx_expr=ctx_expr
                ),
            )
            stmts.append(stmt)
        final = Code(
            "return {$[, ]parts}",
            parts=[
                Code(
                    f"$fname: v_{f.name}",
                    fname=f.serialized_name,
                )
                for f in t.fields
            ],
        )
        stmts.append(final)
        return Def(stmts, VE=ValidationError)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[FieldModelizable], obj: dict, ctx: Context, /):
        (orig_t,) = get_args(t)
        t = model(orig_t)
        forbid_extras = not t.allow_extras
        follow = hasattr(ctx, "follow")
        stmts = []
        if forbid_extras:
            stmts.append(f"used = {sum(1 for f in t.fields if f.required)}")
        args = []

        def _extract(f):
            n = f.name
            if f.metavar:
                return [Code(f"v_{n} = $meta", meta=Code(f.metavar))]

            ctx_expr = (
                Code("$ctx.follow($objt, $obj, $fld)", objt=orig_t, fld=f.name)
                if follow
                else Code("$ctx")
            )

            try_stmts = [Code(f"x_{n} = $obj[$pname]", pname=f.serialized_name)]
            msg = f"Missing required field '{f.serialized_name}' for type `{clsstring(t)}`"
            exc_stmts = [Code("raise $VE($msg)", msg=msg, ve=ValidationError)]
            else_stmts = [
                Code(
                    f"v_{n} = $expr",
                    expr=cls.subcode(
                        "deserialize",
                        f.type,
                        f"x_{n}",
                        ctx,
                        ctx_expr=ctx_expr,
                    ),
                )
            ]

            if f.default is not MISSING:
                if forbid_extras:
                    try_stmts.append("used += 1")
                exc_stmts = [Code(f"v_{n} = $dflt", dflt=f.default)]

            elif f.default_factory is not MISSING:
                if forbid_extras:
                    try_stmts.append("used += 1")
                exc_stmts = [Code(f"v_{n} = $dflt()", dflt=f.default_factory)]

            return [
                "try:",
                try_stmts,
                "except KeyError:",
                exc_stmts,
                "else:",
                else_stmts,
            ]

        def _sortkey(f):
            return an if isinstance(an := f.argument_name, int) else math.inf

        for f in sorted(t.fields, key=_sortkey):
            stmts.extend(_extract(f))
            if isinstance(f.argument_name, str):
                arg = f"{f.argument_name}=v_{f.name}"
            else:
                arg = f"v_{f.name}"
            args.append(arg)

        if forbid_extras:
            stmts.append(
                Code(
                    [
                        "if used != len($obj):",
                        [
                            "extra = set($obj.keys()) - $expected",
                            f"raise $VE(f'Extra unrecognized fields were found for type `{clsstring(t)}`: {{extra}}', ctx=$ctx)",
                        ],
                    ],
                    VE=ValidationError,
                    expected={f.serialized_name for f in t.fields},
                )
            )

        final = Code(
            "return $constructor($[, ]parts)",
            constructor=t.constructor,
            parts=[Code(a) for a in args],
        )
        stmts.append(final)
        return Def(stmts, VE=ValidationError)

    ######################################
    # Implementations: StringModelizable #
    ######################################

    @code_generator(priority=STD2)
    def serialize(self, t: type[StringModelizable], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        m = model(t)
        if not m.accepts(obj) or m.to_string is None:
            return None
        if isinstance(m.to_string, Function):
            return m.to_string
        else:
            return Lambda("$to_string($obj)", to_string=m.to_string)

    @code_generator(priority=STD2)
    def deserialize(self, t: type[StringModelizable], obj: str, ctx: Context, /):
        (t,) = get_args(t)
        m = model(t)
        if m.regexp:
            if isinstance(m.from_string, Def):  # pragma: no cover
                raise Exception("In model definitions, use Lambda with regexp, not Def")
            elif isinstance(m.from_string, Lambda):
                expr = m.from_string.create_expression(["t", "obj", "ctx"])
            else:
                expr = Code("$from_string($obj)", from_string=m.from_string)
            descr = m.string_description or f"pattern {m.regexp.pattern!r}"
            pattern = f"String {{$obj!r}} is not a valid {clsstring(t)}. It should match: {descr}"
            return Def(
                Code(
                    [
                        ["if $regexp.match($obj):", ["return $expr"]],
                        [
                            "else:",
                            [f"""raise $VE(f"{pattern}")"""],
                        ],
                    ]
                ),
                from_string=m.from_string,
                regexp=m.regexp,
                expr=expr,
                descr=descr,
                VE=ValidationError,
            )
        elif isinstance(m.from_string, Function):
            return m.from_string
        else:
            return Lambda("$from_string($obj)", from_string=m.from_string)

    ####################################
    # Implementations: ListModelizable #
    ####################################

    @classmethod
    def __generic_codegen_list(cls, method, m, obj, ctx):
        builder = list if method == "serialize" else m.list_constructor
        lt = m.element_field.type
        comp = "$lbody for IDX, X in enumerate($obj)"
        if builder is list:
            code = f"[{comp}]"
        elif builder is set:
            code = f"{{{comp}}}"
        else:
            code = f"$builder({comp})"
        if hasattr(ctx, "follow"):
            ctx_expr = Code("$ctx.follow($objt, $obj, IDX)", objt=m.original_type)
        else:
            ctx_expr = Code("$ctx")
        return Lambda(
            code,
            lbody=cls.subcode(method, lt, "X", ctx, ctx_expr=ctx_expr),
            builder=builder,
        )

    @code_generator(priority=STD)
    def serialize(cls, t: type[ListModelizable], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        m = model(t)
        return cls.__generic_codegen_list("serialize", m, obj, ctx)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[ListModelizable], obj: list, ctx: Context, /):
        (t,) = get_args(t)
        m = model(t)
        return cls.__generic_codegen_list("deserialize", m, obj, ctx)

    ################################
    # Implementations: Modelizable #
    ################################

    @ovld(priority=STD)
    def schema(self, t: type[Modelizable], ctx: Context, /):
        m = model(t)

        f_schema = s_schema = l_schema = None
        follow = hasattr(ctx, "follow")

        if m.fields is not None:
            properties = {}
            required = []

            for f in m.fields:
                fctx = ctx.follow(t, None, f.name) if follow else ctx
                fsch = recurse(f.type, fctx)
                extra = {}
                if f.description:
                    extra["description"] = f.description
                if f.default is not MISSING:
                    extra["default"] = f.default
                fsch = fsch if not f.description else AnnotatedSchema(fsch, **extra)
                properties[f.serialized_name] = fsch
                if f.required:
                    required.append(f.serialized_name)

            f_schema = {
                "type": "object",
                "properties": properties,
                "required": required,
                "additionalProperties": m.allow_extras,
            }

        if m.from_string is not None:
            s_schema = {"type": "string"}
            if m.regexp:
                s_schema["pattern"] = m.regexp.pattern

        if m.element_field is not None:
            fctx = ctx.follow(t, None, "*") if follow else ctx
            l_schema = {
                "type": "array",
                "items": recurse(m.element_field.type, fctx),
            }

        assert f_schema or s_schema or l_schema
        possibilities = [x for x in [f_schema, s_schema, l_schema] if x]
        match possibilities:
            case [sch]:
                return sch
            case _:
                return {"oneOf": possibilities}

    ###########################
    # Implementations: Unions #
    ###########################

    @code_generator(priority=STD)
    def serialize(cls, t: type[UnionAlias], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        o1, *rest = get_args(t)
        code = cls.subcode("serialize", o1, "$obj", ctx, validate=False)
        for opt in rest:
            code = Code(
                "$ocode if isinstance($obj, $sopt) else $code",
                sopt=basic_type(opt),
                ocode=cls.subcode("serialize", opt, "$obj", ctx, validate=False),
                code=code,
            )
        return Lambda(code)

    @code_generator(priority=STD)
    def deserialize(cls, t: type[UnionAlias] | type[UnionType], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        options = get_args(t)

        tells = [(o, tls) for o in options if (tls := get_tells(opt := o, obj)) is not None]
        if not tells:
            return None

        elim = set()
        for (_, tl1), (_, tl2) in pairwise(tells):
            elim |= tl1 & tl2
        for _, tls in tells:
            tls -= elim

        # TODO: fix this behavior; test it on datetime|None, that's a failure case with
        # this code
        # if len(tells) == 1:
        #     [[opt, tls]] = tells
        #     fn = cls.deserialize.resolve(type[opt], All[obj], ctx)
        #     if getattr(fn, "__codegen__", None):
        #         # TODO: I'm not sure this is correct, because the first argument
        #         # that will actually be given is t, not opt
        #         return fn

        if sum(not tl for _, tl in tells) > 1:
            raise SchemaError(f"Cannot differentiate the possible union members in type '{t}'")

        tells.sort(key=lambda x: len(x[1]))

        (o1, _), *rest = tells

        code = cls.subcode("deserialize", o1, "$obj", ctx)
        for opt, tls in rest:
            code = Code(
                "($ocode if $cond else $code)",
                cond=min(tls).gen(Code("$obj")),
                code=code,
                ocode=cls.subcode("deserialize", opt, "$obj", ctx),
            )
        return Lambda(code)

    @ovld(priority=STD)
    def schema(self, t: type[UnionAlias], ctx: Context, /):
        options = get_args(t)
        return {"oneOf": [recurse(opt, ctx) for opt in options]}

    #########################
    # Implementations: JSON #
    #########################

    @ovld(priority=STD)
    def serialize(self, t: type[Exactly[JSON]], obj: object, ctx: Context, /):
        if isinstance(obj, dict):
            obj = {recurse(str, k, ctx): recurse(JSON, v, ctx) for k, v in obj.items()}
        elif isinstance(obj, list):
            obj = [recurse(JSON, v, ctx) for v in obj]
        elif not isinstance(obj, JSON):
            raise ValidationError(f"Object {obj!r} is not valid JSON", ctx=ctx)
        return obj

    @ovld(priority=STD)
    def deserialize(self, t: type[Exactly[JSON]], obj: object, ctx: Context, /):
        if isinstance(obj, dict):
            obj = {recurse(str, k, ctx): recurse(JSON, v, ctx) for k, v in obj.items()}
        elif isinstance(obj, list):
            obj = [recurse(JSON, v, ctx) for v in obj]
        elif not isinstance(obj, JSON):
            raise ValidationError(f"Object {obj!r} is not valid JSON", ctx=ctx)
        return obj

    @ovld(priority=STD)
    def schema(self, t: type[Exactly[JSON]], ctx: Context, /):
        return {}

    ##########################
    # Implementations: Enums #
    ##########################

    @code_generator(priority=STD3)
    def serialize(cls, t: type[Enum], obj: Enum, ctx: Context, /):
        return Lambda(Code("$obj.value"))

    @code_generator(priority=STD3)
    def deserialize(cls, t: type[Enum], obj: Any, ctx: Context, /):
        (t,) = get_args(t)
        return Lambda(Code("$t($obj)", t=t))

    @ovld(priority=STD3)
    def schema(self, t: type[Enum], ctx: Context, /):
        return {"enum": [e.value for e in t]}

    ##################################
    # Implementations: Literal Enums #
    ##################################

    @ovld(priority=STD)
    def serialize(self, t: type[IsLiteral], obj: Any, ctx: Context, /):
        options = get_args(t)
        if obj not in options:
            raise ValidationError(f"'{obj}' is not a valid option for {t}", ctx=ctx)
        return obj

    @ovld(priority=STD)
    def deserialize(self, t: type[IsLiteral], obj: Any, ctx: Context, /):
        options = get_args(t)
        if obj not in options:
            raise ValidationError(f"'{obj}' is not a valid option for {t}", ctx=ctx)
        return obj

    @ovld(priority=STD)
    def schema(self, t: type[IsLiteral], ctx: Context, /):
        return {"enum": list(get_args(t))}

    ##########################
    # Implementations: Dates #
    ##########################

    @code_generator(priority=STD)
    def deserialize(cls, t: type[datetime], obj: int | float, ctx: Context, /):
        return Lambda(Code("$fromtimestamp($obj)", fromtimestamp=datetime.fromtimestamp))

    # We specify schemas explicitly because they have special formats
    # The serialization/deserializable is taken care of by their model()

    @ovld(priority=STD)
    def schema(self, t: type[date], ctx: Context, /):
        return {"type": "string", "format": "date"}

    @ovld(priority=STD)
    def schema(self, t: type[datetime], ctx: Context, /):
        return {"type": "string", "format": "date-time"}

    #########################
    # Implementations: Path #
    #########################

    @ovld(priority=STD)
    def serialize(self, t: type[Path], obj: Path, ctx: Context):
        if isinstance(ctx, WorkingDirectory) and not obj.is_absolute():
            obj = obj.relative_to(ctx.directory)
        return str(obj)

    @ovld(priority=STD)
    def deserialize(self, t: type[Path], obj: str, ctx: Context):
        pth = Path(obj).expanduser()
        if isinstance(ctx, WorkingDirectory):
            pth = ctx.directory / pth
        return pth

    @ovld(priority=STD)
    def schema(self, t: type[Path], ctx: Context, /):
        return {"type": "string"}

    ##################################
    # Implementations: From __init__ #
    ##################################

    @ovld(priority=LO4)
    def deserialize(self, t: type[object], obj: dict, ctx: Context):
        ot = get_origin(t) or t
        if isinstance(ot.__init__, WrapperDescriptorType):
            msg = f"Cannot deserialize `{obj}` to type `{clsstring(t)}.`"
            if getattr(t, "__annotations__", None) and not is_dataclass(t):
                msg += f" `{clsstring(t)}` appears to have type annotations on some fields, but it is not a dataclass. Did you mean for it to be a dataclass?"
            raise ValidationError(msg, ctx=ctx)
        # We use the Auto feature to convert __init__ to a model. force=True means we
        # will not try to find a model for t itself.
        return recurse(Annotated[t, Auto(call=False, embed_self=False, force=True)], obj, ctx)

    @ovld(priority=LO4)
    def schema(self, t: type[object], ctx: Context):
        if t.__init__ is object.__init__:
            msg = f"Cannot generate schema for type `{clsstring(t)}`."
            if getattr(t, "__annotations__", None) and not is_dataclass(t):
                msg += f" `{clsstring(t)}` appears to have type annotations on some fields, but it is not a dataclass. Did you mean for it to be a dataclass?"
            raise ValidationError(msg, ctx=ctx)
        return recurse(Annotated[t, Auto(call=False, embed_self=False, force=True)], ctx)



import importlib
from json import JSONEncoder

from ovld import Medley, recurse
from ovld.dependent import HasKey

from serieux import Context, Serieux, deserialize

##################
# Implementation #
##################


@Serieux.extend
class Target(Medley):
    def deserialize(self, t: type[object], obj: HasKey["_target_"], ctx: Context):  # noqa
        obj = dict(obj)
        symbol = obj.pop("_target_")
        module, symbol = symbol.split(".", 1)
        target = getattr(importlib.import_module(module), symbol)
        kwargs = {k: recurse(type(v), v, ctx) for k, v in obj.items()}
        return target(**kwargs)


#################
# Demonstration #
#################


def main():
    data = {
        "_target_": "json.JSONEncoder",
        "indent": 3,
    }
    result = deserialize(object, data)
    assert isinstance(result, JSONEncoder)
    print(result.encode({"x": 3, "y": 4}))


if __name__ == "__main__":
    main()



class PriorityLevel(tuple):
    def __new__(cls, *elements):
        obj = super().__new__(cls, elements)
        obj.current = 0
        return obj

    def next(self):
        self.current += 1
        return PriorityLevel(*self, self.current)

    def __call__(self, *elems):
        return PriorityLevel(*self, *elems)


class NamedInteger(int):
    def __new__(cls, value, name):
        obj = super().__new__(cls, value)
        obj._name = name
        return obj

    def __str__(self):
        return self._name

    __repr__ = __str__


MAX = PriorityLevel(NamedInteger(100, "MAX"))
HI9 = PriorityLevel(NamedInteger(9, "HI9"))
HI8 = PriorityLevel(NamedInteger(8, "HI8"))
HI7 = PriorityLevel(NamedInteger(7, "HI7"))
HI6 = PriorityLevel(NamedInteger(6, "HI6"))
HI5 = PriorityLevel(NamedInteger(5, "HI5"))
HI4 = PriorityLevel(NamedInteger(4, "HI4"))
HI3 = PriorityLevel(NamedInteger(3, "HI3"))
HI2 = PriorityLevel(NamedInteger(2, "HI2"))
HIGH = HI1 = PriorityLevel(NamedInteger(1, "HI1"))
DEFAULT = PriorityLevel(NamedInteger(0, "DEFAULT"))
STD5 = PriorityLevel(NamedInteger(-1, "STD5"))
STD4 = PriorityLevel(NamedInteger(-2, "STD4"))
STD3 = PriorityLevel(NamedInteger(-3, "STD3"))
STD2 = PriorityLevel(NamedInteger(-4, "STD2"))
STD = STD1 = PriorityLevel(NamedInteger(-5, "STD1"))
LOW = LO1 = PriorityLevel(NamedInteger(-11, "LO1"))
LO2 = PriorityLevel(NamedInteger(-12, "LO2"))
LO3 = PriorityLevel(NamedInteger(-13, "LO3"))
LO4 = PriorityLevel(NamedInteger(-14, "LO4"))
LO5 = PriorityLevel(NamedInteger(-15, "LO5"))
MIN = PriorityLevel(NamedInteger(-100, "MIN"))



from __future__ import annotations

from dataclasses import dataclass

import pytest
from ovld import Dataclass, Dependent, Lambda, ovld
from ovld.dependent import Regexp

from serieux import deserialize, schema, serialize
from serieux.features.clargs import CommandLineArguments
from serieux.model import Field, Model


def _rgb_from_string(obj, cls):
    hex_str = obj.lstrip("#")
    red = int(hex_str[0:2], 16)
    green = int(hex_str[2:4], 16)
    blue = int(hex_str[4:6], 16)
    return cls(red=red, green=green, blue=blue)


@dataclass
class RGB:
    red: int
    green: int
    blue: int

    @classmethod
    def serieux_deserialize(cls, obj, ctx, call_next):
        if isinstance(obj, str):
            return _rgb_from_string(obj, cls=RGB)
        else:
            return call_next(cls, obj, ctx)

    @classmethod
    def serieux_serialize(cls, obj, ctx, call_next):
        if 0 <= obj.red <= 255:
            return f"#{obj.red:02x}{obj.green:02x}{obj.blue:02x}"
        else:
            return call_next(cls, obj, ctx)

    @classmethod
    def serieux_schema(cls, ctx, call_next):
        return {
            "oneOf": [
                {"type": "string", "pattern": r"^#[0-9a-fA-F]{6}$"},
                call_next(cls, ctx),
            ]
        }


def test_custom_deserialize():
    assert deserialize(RGB, "#ff00ff") == RGB(red=255, green=0, blue=255)
    assert deserialize(RGB, {"red": 255, "green": 100, "blue": 100}) == RGB(
        red=255, green=100, blue=100
    )


def test_custom_serialize():
    assert serialize(RGB, RGB(red=255, green=0, blue=255)) == "#ff00ff"
    assert serialize(RGB, RGB(red=1000, green=0, blue=0)) == {"red": 1000, "green": 0, "blue": 0}


def test_custom_schema():
    assert schema(RGB).compile(root=False) == {
        "oneOf": [
            {"type": "string", "pattern": r"^#[0-9a-fA-F]{6}$"},
            {
                "type": "object",
                "properties": {
                    "red": {"type": "integer"},
                    "green": {"type": "integer"},
                    "blue": {"type": "integer"},
                },
                "required": ["red", "green", "blue"],
                "additionalProperties": False,
            },
        ]
    }


@dataclass
class RGBO:
    red: int
    green: int
    blue: int

    @classmethod
    @ovld
    def serieux_deserialize(cls, obj: Regexp[r"^#[0-9a-fA-F]{6}$"], ctx, call_next):
        return _rgb_from_string(obj, cls=RGBO)

    @classmethod
    @ovld
    def serieux_serialize(
        cls, obj: Dependent[Dataclass, lambda rgb: 0 <= rgb.red <= 255], ctx, call_next
    ):
        assert 0 <= obj.red <= 255
        return f"#{obj.red:02x}{obj.green:02x}{obj.blue:02x}"


def test_custom_deserialize_o():
    assert deserialize(RGBO, "#ff00ff") == RGBO(red=255, green=0, blue=255)
    assert deserialize(RGBO, {"red": 255, "green": 100, "blue": 100}) == RGBO(
        red=255, green=100, blue=100
    )


def test_custom_serialize_o():
    assert serialize(RGBO, RGBO(red=255, green=0, blue=255)) == "#ff00ff"
    assert serialize(RGBO, RGBO(red=1000, green=0, blue=0)) == {"red": 1000, "green": 0, "blue": 0}


class RGBM:
    red: int
    green: int
    blue: int

    def __init__(self, red, green, blue):
        self.red = red
        self.green = green
        self.blue = blue

    @classmethod
    def serieux_model(cls, call_next):
        return Model(
            original_type=cls,
            fields=[
                Field(name="red", type=int, serialized_name="R"),
                Field(name="green", type=int, serialized_name="G"),
                Field(name="blue", type=int, serialized_name="B"),
            ],
            constructor=cls,
            from_string=Lambda("$from_string($obj, $t)", from_string=_rgb_from_string),
            regexp=r"^#[0-9a-fA-F]{6}$",
            string_description="A RGB color in hex, such as #ff0000",
        )


def test_custom_deserialize_m():
    obj = deserialize(RGBM, {"R": 30, "G": 100, "B": 200})
    assert isinstance(obj, RGBM)
    assert obj.red == 30 and obj.green == 100 and obj.blue == 200


def test_custom_deserialize_m_from_string():
    obj = deserialize(RGBM, "#ff0000")
    assert isinstance(obj, RGBM)
    assert obj.red == 255 and obj.green == 0 and obj.blue == 0


def test_custom_m_schema(file_regression):
    assert schema(RGBM).compile(root=False) == {
        "oneOf": [
            {
                "type": "object",
                "properties": {
                    "R": {"type": "integer"},
                    "G": {"type": "integer"},
                    "B": {"type": "integer"},
                },
                "required": ["R", "G", "B"],
                "additionalProperties": False,
            },
            {"type": "string", "pattern": "^#[0-9a-fA-F]{6}$"},
        ]
    }


@dataclass
class HasColor:
    color: RGBM


def test_clargs_rgbm():
    clargs = CommandLineArguments(
        arguments=["--color", "#ff0000"],
        mapping={"": {"auto": True}},
    )
    obj = deserialize(HasColor, clargs)
    color = obj.color
    assert isinstance(color, RGBM)
    assert color.red == 255 and color.green == 0 and color.blue == 0


def test_clargs_bad_regexp():
    clargs = CommandLineArguments(
        arguments=["--color", "#ff00XX"],
        mapping={"": {"auto": True}},
    )
    with pytest.raises(SystemExit):
        deserialize(HasColor, clargs)


@dataclass
class RGBS:
    red: int
    green: int
    blue: int

    @classmethod
    def serieux_from_string(cls, s):
        return _rgb_from_string(s, cls)

    @classmethod
    def serieux_to_string(cls, obj):
        return f"#{obj.red:02x}{obj.green:02x}{obj.blue:02x}"


def test_serieux_from_string_and_to_string_dataclass():
    rgb = deserialize(RGBS, "#ff00ff")
    assert rgb == RGBS(red=255, green=0, blue=255)

    s = serialize(RGBS, rgb)
    assert s == "#ff00ff"

    sch = schema(RGBS).compile(root=False)
    assert sch == {
        "oneOf": [
            {
                "type": "object",
                "properties": {
                    "red": {"type": "integer"},
                    "green": {"type": "integer"},
                    "blue": {"type": "integer"},
                },
                "required": ["red", "green", "blue"],
                "additionalProperties": False,
            },
            {"type": "string"},
        ]
    }


class RGBS2:
    def __init__(self, red: int, green: int, blue: int):
        self.red = red
        self.green = green
        self.blue = blue

    @classmethod
    def serieux_from_string(cls, s):
        return _rgb_from_string(s, cls)

    @classmethod
    def serieux_to_string(cls, obj):
        return f"#{obj.red:02x}{obj.green:02x}{obj.blue:02x}"


def test_serieux_from_string_and_to_string():
    rgb = deserialize(RGBS2, "#ff00ff")
    assert isinstance(rgb, RGBS2)
    assert (rgb.red, rgb.green, rgb.blue) == (255, 0, 255)

    s = serialize(RGBS2, rgb)
    assert s == "#ff00ff"

    sch = schema(RGBS2).compile(root=False)
    assert sch == {"type": "string"}



import re
import sys
from dataclasses import dataclass, field

from ovld import ovld
from ovld.medley import ABSENT

from .ctx import Context, Location, Trail, locate
from .formats import FileSource


@dataclass
class ContextInformation:
    func: str = None
    path: list[str] = field(default_factory=list)
    locs: list[Location] = field(default_factory=list)
    ctx: Context = None

    @property
    def trail_string(self):
        return "".join([f".{field}" for field in self.path]) if self.path else "(at root)"


def _color(code, text):
    return f"\u001b[1m\u001b[{code}m{text}\u001b[0m"


@ovld
def find_link(o1: dict, o2):
    for k, v in o1.items():
        if v is o2:
            return k
    return None


@ovld
def find_link(o1: list, o2):
    for i, v in enumerate(o1):
        if v is o2:
            return str(i)
    return None


@ovld
def find_link(o1, o2):
    try:
        v1 = vars(o1)
    except TypeError:
        return None
    return find_link(v1, o2)


def extract_information(ctx=None, frame=None):
    frame = frame or sys._getframe(1)
    ci = ContextInformation(ctx=ctx)

    def _fast_path():
        ci.func = ci.func or "<serieux>"
        ci.path[:0] = ci.ctx.trail
        if loc := locate(ci.ctx):
            ci.locs.insert(0, loc)
        return ci

    if isinstance(ci.ctx, Trail):
        return _fast_path()

    above = None
    while frame:
        lcls = frame.f_locals
        ci.ctx = lcls.get("ctx", ci.ctx)
        if isinstance(ci.ctx, Trail):
            return _fast_path()
        elif ci.ctx and (m := re.match(r"^((?:de)?serialize|schema)\[", frame.f_code.co_name)):
            ci.func = m.groups()[0]
            t1, obj1 = lcls.get("t", ABSENT), lcls.get("obj", ABSENT)
            obj2 = None
            if above is not None:
                _, obj2 = above
                if lnk := find_link(obj1, obj2):
                    ci.path.insert(0, lnk)
            if obj1 is not obj2 and isinstance(obj1, FileSource):
                try:
                    fp = obj1.field.split(".") if obj1.field else []
                    loc = obj1.format.locate(obj1.path, [*fp, *ci.path])
                    ci.locs.insert(0, loc)
                except FileNotFoundError:
                    pass
            above = t1, obj1
        frame = frame.f_back
    return ci


def display_context_information(
    template="An error happened in serieux.{func} at location {trail}",
    *,
    ctx=None,
    exc=None,
    frame=None,
    show_source=True,
    file=sys.stderr,
    message="ERROR",
    **kwargs,
):
    ci = None
    if exc is not None:
        if isinstance(exc, IndividualSerieuxError) and exc.ctx:
            ctx = exc.ctx
            ci = exc.info
        elif isinstance(exc, ValidationExceptionGroup):  # pragma: no cover
            display(exc)
            return
        else:
            tb = exc.__traceback__
            while tb:
                frame = tb.tb_frame
                tb = tb.tb_next
    ci = ci or extract_information(ctx, frame)
    if ci.func is None:  # pragma: no cover
        return
    print(
        template.format(trail=_color(33, ci.trail_string), func=ci.func, message=message),
        file=file,
    )
    if show_source and ci.locs:
        for location in ci.locs:
            display_location(location, file=file, **kwargs)


def display_location(location, source_context=1, indent=0, ellipsis_cutoff=3, file=sys.stderr):
    width = 3 + indent
    (l1, c1), (l2, c2) = location.linecols
    if c2 == 0:
        l2 -= 1
        c2 = 10_000_000_000_000
    lines = location.whole_text.split("\n")
    start = l1 - source_context
    while start < 0 or not lines[start].strip():
        start += 1
    end = l2 + source_context
    while end >= len(lines) or not lines[end].strip():
        end -= 1

    print(f"{'':{indent}}@ {location.source.absolute()}:{l1 + 1}", file=file)
    for li in range(start, end + 1):
        line = lines[li]
        if li == l2 and not line.strip():  # pragma: no cover
            break
        if li == l1 + ellipsis_cutoff and li < l2:
            print(f"{'':{width}}  ...", file=file)
            continue
        elif li > l1 + ellipsis_cutoff and li < l2:
            continue

        hls = hle = 0
        if li == l1:
            hls = c1
        if li >= l1 and li < l2:
            hle = len(line)
        if li == l2:
            hle = c2

        if hls or hle:
            line = line[:hls] + _color(31, line[hls:hle]) + line[hle:]

        print(f"{li + 1:{width}}: {line}", file=file)


def merge_errors(*errors):
    collected = []
    for err in errors:
        if isinstance(err, ExceptionGroup):
            collected.extend(err.exceptions)
        elif isinstance(err, Exception):
            collected.append(err)
    return ValidationExceptionGroup("Some errors occurred", collected) if collected else None


class SerieuxError(Exception):
    pass


class IndividualSerieuxError(SerieuxError):
    def __init__(self, message=None, *, ctx=None):
        super().__init__(message)
        self.info = extract_information(ctx)
        self.ctx = self.info.ctx

    @property
    def message(self):
        return self.args[0]

    def __str__(self):
        if location := locate(self.ctx):
            (l1, c1), (l2, c2) = location.linecols
            lc = f"{l1}:{c1}-{l2}:{c2}" if l1 != l2 else f"{l1}:{c1}-{c2}"
            return f"{location.source}:{lc} -- {self.message}"
        else:
            return f"At path {self.info.trail_string}: {self.message}"


class NotGivenError(IndividualSerieuxError):
    pass


class ValidationExceptionGroup(SerieuxError, ExceptionGroup):
    def derive(self, excs):  # pragma: no cover
        return ValidationExceptionGroup(self.message, excs)


class ValidationError(IndividualSerieuxError):
    def __init__(self, message=None, *, exc=None, ctx=None):
        if message is None:
            message = f"{type(exc).__name__}: {exc}"
        super().__init__(message=message, ctx=ctx)
        self.exc = exc


class SchemaError(IndividualSerieuxError):
    def __init__(self, message=None, *, exc=None, ctx=None):
        if message is None:  # pragma: no cover
            message = f"{type(exc).__name__}: {exc}"
        super().__init__(message=message, ctx=ctx)
        self.exc = exc


def display(exc, file=sys.stderr):
    if isinstance(exc, ValidationExceptionGroup):
        for i, subexc in enumerate(exc.exceptions):
            print(f"[#{i}] ", end="", file=file)
            display(subexc, file)
    else:
        if isinstance(exc, IndividualSerieuxError):
            msg = exc.message
        else:
            msg = f"{type(exc).__name__}: {exc}"
        display_context_information(
            "{trail}: {message}",
            exc=exc,
            file=file,
            message=msg,
            indent=2,
        )



from __future__ import annotations

from dataclasses import dataclass

from .definitions import Point


@dataclass
class Tree[T]:
    left: Tree[T] | T
    right: Tree[T] | T


type ListTreeP = list[Point | ListTreeP]


@dataclass
class LTHolder:
    lt: ListTreeP



from dataclasses import dataclass

import cryptography
import pytest

from serieux.__main__ import (
    Check,
    Dump,
    Patch,
    Run,
    Schema,
    model_at,
    value_at,
)
from serieux.features.encrypt import crypt_prefix
from tests.definitions import Pig, Point, World
from tests.features.test_encrypt import User


def test_value_at_dict():
    data = {"a": {"b": {"c": 42}}}
    assert value_at(data, "a.b.c") == 42


def test_value_at_list():
    data = {"items": [{"name": "first"}, {"name": "second"}]}
    assert value_at(data, "items.1.name") == "second"


def test_value_at_object():
    point = Point(x=10, y=20)
    data = {"point": point}
    assert value_at(data, "point.x") == 10
    assert value_at(data, "point.y") == 20


def test_model_at():
    assert model_at(Point, "x") is int
    assert model_at(list[Point], "0.x") is int


def test_dump_command(capsys, datapath, file_regression):
    dumper = Dump(
        model=World,
        file=datapath / "world.yaml",
        format="yaml",
        password="NOT_USED",
    )
    dumper()
    captured = capsys.readouterr()
    file_regression.check(captured.out)


def test_dump_command_out(tmp_path, datapath, file_regression):
    out_file = tmp_path / "dump.yaml"
    dumper = Dump(
        model=World,
        file=datapath / "world.yaml",
        format="yaml",
        password="NOT_USED",
        out=out_file,
    )
    dumper()
    assert out_file.exists()
    content = out_file.read_text()
    file_regression.check(content)


def test_dump_command_encrypted(tmp_path, datapath):
    out_file = tmp_path / "dump.yaml"
    dumper = Dump(
        model=list[User],
        file=datapath / "users-encrypted.yaml",
        format="yaml",
        password="gaga",
        out=out_file,
    )
    dumper()
    content = out_file.read_text()
    assert "super_secret" in content
    assert "abc123" in content
    assert "letmein" in content


def test_dump_command_encrypted_env_password(tmp_path, datapath, monkeypatch):
    out_file = tmp_path / "dump.yaml"
    monkeypatch.setenv("SERIEUX_PASSWORD", "gaga")
    dumper = Dump(
        model=list[User],
        file=datapath / "users-encrypted.yaml",
        format="yaml",
        out=out_file,
    )
    dumper()
    content = out_file.read_text()
    assert "super_secret" in content
    assert "abc123" in content
    assert "letmein" in content


def test_dump_command_wrong_password(datapath):
    dumper = Dump(
        model=list[User],
        file=datapath / "users-encrypted.yaml",
        format="yaml",
        password="WRONG",
    )
    with pytest.raises(cryptography.fernet.InvalidToken):
        dumper()


def test_check_command(datapath):
    for sel, code in [("weight", 0), ("beautiful", 1), ("inexistent", 2)]:
        checker = Check(
            model=Pig,
            file=datapath / "beastfly.yaml",
            select=sel,
            password="NOT_USED",
        )

        with pytest.raises(SystemExit) as exc:
            checker()
        assert exc.value.code == code


def test_schema_command(capsys, file_regression):
    schema_cmd = Schema(model=World)
    schema_cmd()
    captured = capsys.readouterr()
    file_regression.check(captured.out)


def test_schema_command_out(tmp_path, file_regression):
    out_file = tmp_path / "schema.json"
    schema_cmd = Schema(model=World, out=out_file)
    schema_cmd()
    assert out_file.exists()
    content = out_file.read_text()
    file_regression.check(content)


def test_patch_command(tmp_path, datapath):
    out_file = tmp_path / "encrypted.yaml"
    patcher = Patch(
        model=list[User],
        file=datapath / "users.yaml",
        password="gaga",
        out=out_file,
    )
    patcher()
    assert out_file.exists()
    content = out_file.read_text()

    assert "super_secret" not in content
    assert "abc123" not in content
    assert "letmein" not in content

    assert crypt_prefix in content


def test_patch_command_inplace(tmp_path, datapath):
    out_file = tmp_path / "encrypted.yaml"
    out_file.write_bytes((datapath / "users.yaml").read_bytes())

    content = out_file.read_text()
    assert "super_secret" in content
    assert "abc123" in content
    assert "letmein" in content
    assert crypt_prefix not in content

    patcher = Patch(
        model=list[User],
        file=out_file,
        password="gaga",
        out=out_file,
    )
    patcher()

    content = out_file.read_text()
    assert "super_secret" not in content
    assert "abc123" not in content
    assert "letmein" not in content
    assert crypt_prefix in content


def addnums(x: int, y: int, /):
    return x + y


def test_run_command(capsys):
    runner = Run(
        func=addnums,
        args=["6", "7"],
    )
    runner()
    captured = capsys.readouterr()
    assert captured.out.strip() == "13"


@dataclass
class Multiplier:
    left: int
    right: int

    def __call__(self):
        return self.left * self.right


def test_run_command_dataclass(capsys):
    runner = Run(
        func=Multiplier,
        args=["--left", "6", "--right", "7"],
    )
    runner()
    captured = capsys.readouterr()
    assert captured.out.strip() == "42"



from __future__ import annotations

from dataclasses import dataclass, field
from datetime import date
from enum import Enum
from numbers import Number
from pathlib import Path


@dataclass
class Tree:
    left: Tree | Number
    right: Tree | Number


@dataclass
class Elf:
    name: str
    birthdate: date
    favorite_color: str


@dataclass
class Citizen:
    name: str
    birthyear: int
    hometown: str


@dataclass
class Country:
    languages: list[str]
    capital: str
    population: int
    citizens: list[Citizen]


@dataclass
class World:
    countries: dict[str, Country]


@dataclass
class Point:
    x: int
    y: int


@dataclass
class Point3D(Point):
    z: int


# This one is not a dataclass and useful error messages should point that out
class Pointato:
    x: int
    y: int


class Color(str, Enum):
    RED = "red"
    GREEN = "green"
    BLUE = "blue"


class Level(Enum):
    HI = 2
    MED = 1
    LO = 0


@dataclass
class Pig:
    # How pink the pig is
    pinkness: float

    weight: float
    """Weight of the pig, in kilograms"""

    # Is the pig...
    # truly...
    beautiful: bool = True  # ...beautiful?


@dataclass
class Defaults:
    name: str
    aliases: list[str] = field(default_factory=list)
    cool: bool = field(default=False, kw_only=True)


@dataclass
class Player:
    first: str
    last: str
    batting: float


@dataclass
class Team:
    name: str
    players: list[Player]


@dataclass(frozen=True)
class Job:
    # Name of the job
    title: str
    # How much it pays, in dollars
    yearly_pay: float


@dataclass
class Worker:
    name: str
    job: Job = None


DID = dict[str, "int | DID"]


@dataclass
class DIDHolder:
    did: DID


ListTree = list["Point | ListTree"]


@dataclass
class LTHolder:
    lt: ListTree


class DotDict(dict):
    def __getattr__(self, attr):
        return self[attr]


@dataclass
class Character:
    name: str
    age: int
    occupation: str
    backstory: str

    class SerieuxConfig:
        allow_extras = True


@dataclass(kw_only=True)
class Car:
    horsepower: int


@dataclass
class IdentifiedCar(Car):
    id: int


@dataclass
class File:
    path: Path

    # [ignore]
    fd: object = None

    def __post_init__(self):
        self.fd = open(self.path, "r")



import json
from enum import Enum
from typing import Any, Counter

from ovld import Medley, call_next, recurse

from .features.partial import merge


class Schema:
    def __init__(self, t):
        self.for_type = t
        self.data = {}
        self.redirect = None

    def update(self, data):
        if isinstance(data, Schema):
            data = data.data
        if isinstance(data, dict):
            self.data.update(data)
        else:
            assert not self.data
            self.redirect = data

    def get(self, key, default):
        return self.data.get(key, default)

    def __contains__(self, key):
        return key in self.data

    def compile(self, **kwargs):
        return SchemaCompiler(**kwargs)(self)

    def json(self, fp=None):
        if fp:  # pragma: no cover
            return json.dump(self.compile(), fp, indent=4)
        else:
            return json.dumps(self.compile(), indent=4)


class AnnotatedSchema:
    def __init__(self, parent, **annotations):
        self.parent = parent
        self.data = annotations


class RefPolicy(str, Enum):
    ALWAYS = "always"
    NOREPEAT = "norepeat"
    MINIMAL = "minimal"
    NEVER = "never"


class SchemaCompiler(Medley):
    ref_policy: RefPolicy = RefPolicy.NOREPEAT
    root: bool = True

    def __post_init__(self):
        self.refs = {}
        self.defs = {}
        self.done = set()
        self.name_indexes = Counter()

    def unique_name(self, t: Any):
        name = t.__name__
        idx = self.name_indexes[name]
        self.name_indexes[name] += 1
        if idx > 0:
            name = f"{name}{idx + 1}"
        return name

    def __call__(self, x: object):
        rval = recurse(x, ("#",))
        if self.root:
            rval["$schema"] = "https://json-schema.org/draft/2020-12/schema"
        if self.defs:
            rval["$defs"] = self.defs
        return rval

    def __call__(self, d: dict, pth: tuple):
        return {k: recurse(v, (*pth, k)) for k, v in d.items()}

    def __call__(self, xs: list, pth: tuple):
        return [recurse(x, (*pth, str(i))) for i, x in enumerate(xs)]

    def __call__(self, x: object, pth: tuple):
        return x

    def __call__(self, x: Schema, pth: tuple):
        if x.redirect:
            return recurse(x.redirect, pth)
        is_always = self.ref_policy == RefPolicy.ALWAYS
        if x.get("type", "object") not in ("object", "array") or "oneOf" in x:
            return call_next(x.data, pth)
        elif x in self.refs:
            if x not in self.done and self.ref_policy == RefPolicy.NEVER:
                raise Exception("Recursive schema cannot be compiled without $ref")
            elif x not in self.done or self.ref_policy not in (RefPolicy.NEVER, RefPolicy.MINIMAL):
                return {"$ref": "/".join(self.refs[x])}
            else:
                return call_next(x.data, pth)
        else:
            if is_always:
                name = self.unique_name(x.for_type)
                pth = ("#", "$defs", name)
            self.refs[x] = pth
            rval = call_next(x.data, pth)
            if "$ref" not in rval:
                self.done.add(x)
            if is_always:
                self.defs[name] = rval
                return {"$ref": f"#/$defs/{name}"}
            return rval

    def __call__(self, x: AnnotatedSchema, pth: tuple):
        rval = recurse(x.parent, pth)
        return merge(rval, recurse(x.data, pth))



from ..utils import import_any
from .abc import FileFormat

loads = import_any(
    feature="TOML loading",
    candidates={
        "toml": lambda m: m.loads,
        "tomllib": lambda m: m.loads,
        "tomli": lambda m: m.loads,
        "tomli_w": None,
    },
)

dumps = import_any(
    feature="TOML dumping",
    candidates={
        "toml": lambda m: m.dumps,
        "tomli-w:tomli_w": lambda m: m.dumps,
        "tomli": None,
        "tomllib": None,
    },
)


class TOML(FileFormat):
    def loads(self, s: str):
        return loads(s)

    def dumps(self, data):
        return dumps(data)



import importlib
import importlib.metadata
import logging
import os
import sys
import traceback
from functools import partial
from pathlib import Path
from typing import TYPE_CHECKING, TypeAlias

from .auto import Auto
from .ctx import Context, Patch, Patcher, Trail, WorkingDirectory
from .exc import (
    SerieuxError,
    ValidationError,
    ValidationExceptionGroup,
    display_context_information,
)
from .features.clargs import CLIDefinition, CommandLineArguments, parse_cli
from .features.comment import Comment, CommentRec
from .features.dotted import DottedNotation
from .features.fromfile import IncludeFile
from .features.interpol import Environment
from .features.lazy import DeepLazy, Lazy, LazyProxy
from .features.partial import AllTrails, Partial, Sources
from .features.registered import AutoRegistered, Referenced, auto_singleton
from .features.tagset import ReferencedClass, Tagged, TaggedSubclass, TaggedUnion
from .impl import BaseImplementation
from .instructions import Instruction
from .model import AllowExtras, Field, FieldModelizable, Model, Modelizable, StringModelizable
from .schema import RefPolicy, Schema
from .utils import JSON, check_signature
from .version import version as __version__

logger = logging.getLogger("serieux")


def _default_features():
    features = []
    eps = importlib.metadata.entry_points(group="serieux.default_features")
    for ep in eps:
        try:
            feature = ep.load()
        except ImportError:  # pragma: no cover
            # Some features may be dependent on what packages are installed.
            # That is fine.
            continue
        except Exception:  # pragma: no cover
            logger.warning(
                "Default serieux feature %r (%s) failed to load:\n%s",
                ep.name,
                ep.value,
                traceback.format_exc(),
            )
        else:
            features.append(feature)
    features.sort(key=lambda t: -len(t.mro()))
    return features


default_features = _default_features()


if TYPE_CHECKING:  # pragma: no cover
    from typing import TypeVar

    T = TypeVar("T")

    JSON: TypeAlias = list["JSON"] | dict[str, "JSON"] | int | str | float | bool | None

    class _MC:
        def __add__(self, other) -> type["Serieux"]: ...

    class Serieux(metaclass=_MC):
        def dump(
            self, t: type[T], obj: object, ctx: Context = None, *, dest: Path = None
        ) -> JSON | None: ...

        def load(self, t: type[T], obj: object, ctx: Context = None) -> T: ...

        def serialize(self, t: type[T], obj: object, ctx: Context = None) -> JSON: ...

        def deserialize(self, t: type[T], obj: object, ctx: Context = None) -> T: ...

        def schema(self, t: type[T], ctx: Context = None) -> Schema[str, "JSON"]: ...

        def __add__(self, other) -> "Serieux": ...

else:

    class Serieux(BaseImplementation, *default_features):
        pass


serieux = Serieux()
serialize = serieux.serialize
deserialize = serieux.deserialize
schema = serieux.schema
load = serieux.load
dump = serieux.dump
get_serializer = serieux.get_serializer
get_deserializer = serieux.get_deserializer


def serializer(fn=None, priority=0):
    if fn is None:
        return partial(serializer, priority=priority)

    check_signature(fn, "serializer", ("self", "t: type[T1]", "obj: T2", "ctx: T3>Context"))
    Serieux.serialize.register(fn, priority=priority)
    return fn


def deserializer(fn=None, priority=0):
    if fn is None:
        return partial(deserializer, priority=priority)

    check_signature(fn, "deserializer", ("self", "t: type[T1]", "obj: T2", "ctx: T3>Context"))
    Serieux.deserialize.register(fn, priority=priority)
    return fn


def schema_definition(fn=None, priority=0):
    if fn is None:
        return partial(schema_definition, priority=priority)

    check_signature(fn, "schema definition", ("self", "t: type[T1]", "ctx: T2>Context"))
    Serieux.schema.register(fn, priority=priority)
    return fn


if sys.excepthook is sys.__excepthook__ and not os.getenv("SERIEUX_DISABLE_EXCEPTHOOK", None):

    def custom_excepthook(exc_type, exc_value, exc_traceback):  # pragma: no cover
        sys.__excepthook__(exc_type, exc_value, exc_traceback)
        display_context_information(exc=exc_value)

    sys.excepthook = custom_excepthook


__all__ = [
    "__version__",
    "AllowExtras",
    "AllTrails",
    "Auto",
    "AutoRegistered",
    "auto_singleton",
    "BaseImplementation",
    "CLIDefinition",
    "CommandLineArguments",
    "Comment",
    "CommentRec",
    "Context",
    "DeepLazy",
    "deserialize",
    "display_context_information",
    "DottedNotation",
    "dump",
    "Environment",
    "Field",
    "get_deserializer",
    "get_serializer",
    "IncludeFile",
    "JSON",
    "Lazy",
    "LazyProxy",
    "load",
    "Model",
    "FieldModelizable",
    "Modelizable",
    "Instruction",
    "parse_cli",
    "Partial",
    "Patch",
    "Patcher",
    "RefPolicy",
    "Referenced",
    "ReferencedClass",
    "Registered",
    "schema",
    "Schema",
    "serialize",
    "serieux",
    "Serieux",
    "SerieuxError",
    "Sources",
    "StringModelizable",
    "Tagged",
    "TaggedUnion",
    "TaggedSubclass",
    "Trail",
    "ValidationError",
    "ValidationExceptionGroup",
    "WorkingDirectory",
]



import argparse
import json
import re
import sys
from dataclasses import dataclass, field
from enum import Enum
from types import NoneType
from typing import Annotated, Any, get_args

from ovld import Medley, ovld, recurse

from ..ctx import Context
from ..instructions import pushdown, strip
from ..model import Field, FieldModelizable, ListModelizable, StringModelizable, field_at, model
from ..utils import IsLiteral, UnionAlias, clsstring
from .dotted import unflatten
from .partial import Sources
from .tagset import TagSet, decompose, tag_field

##################
# Implementation #
##################


def _compose(dest, new_part):
    return f"{dest}.{new_part}" if dest else new_part


def _soft_conversion(t):
    def filter(x):
        try:
            return t(x)
        except (ValueError, TypeError):
            return x

    return filter


@dataclass
class CommandLineArguments:
    arguments: list[str]
    mapping: dict[str, str | dict[str, Any]] = field(default_factory=lambda: {"": {"auto": True}})

    def parse(self, root_type, argv):
        return parse_cli(root_type=root_type, mapping=self.mapping, argv=argv)


class ConcatenateAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, " ".join(values) if isinstance(values, list) else values)


class ParseStringAction(argparse.Action):
    def __call__(self, parser, namespace, value, option_string=None):
        if isinstance(value, str) and re.fullmatch(string=value, pattern=r"[\[\{].*[\}\]]"):
            value = json.loads(value)
        setattr(namespace, self.dest, value)


@ovld
def make_argument(t: type[str], partial: dict, model_field: Field):
    rval = {**partial, "type": t}
    if "nargs" in partial and "action" not in partial:
        rval["action"] = ConcatenateAction
    return rval


@ovld
def make_argument(t: type[int] | type[float], partial: dict, model_field: Field):
    return {**partial, "type": _soft_conversion(t)}


@ovld
def make_argument(t: type[bool], partial: dict, model_field: Field):
    partial.pop("metavar", None)
    partial.pop("type", None)
    return {**partial, "action": argparse.BooleanOptionalAction}


@ovld
def make_argument(t: type[ListModelizable], partial: dict, model_field: Field):
    lt = model(t).element_field.type
    if partial.get("action", None) == "append":
        return {"type": lt, **partial}
    else:
        return {"nargs": "*", "type": _soft_conversion(lt), **partial}


@ovld(priority=1)
def make_argument(t: type[Enum], partial: dict, model_field: Field):
    return {**partial, "type": str, "choices": [e.value for e in t]}


@ovld
def make_argument(t: type[IsLiteral], partial: dict, model_field: Field):
    return {**partial, "type": str, "choices": [x for x in t.__args__]}


@ovld(priority=-1)  # pragma: no cover
def make_argument(t: type[Annotated], partial: dict, model_field: Field):
    return recurse(pushdown(t), partial, model_field)


def regex_checker(pattern, descr):
    def check(value):
        if not re.fullmatch(pattern, value):
            msg = (
                f"is invalid. It must be: {descr}"
                if descr
                else f"does not match pattern {pattern!r}"
            )
            raise argparse.ArgumentError(None, f"Value {value!r} {msg}")
        return value

    return check


@ovld
def make_argument(t: type[StringModelizable], partial: dict, model_field: Field):
    m = model(t)
    return {**partial, "type": regex_checker(m.regexp, m.string_description) if m.regexp else str}


@ovld(priority=-2)
def make_argument(t: type[object], partial: dict, model_field: Field):
    return {**partial, "action": ParseStringAction}


@ovld(priority=1)
def make_argument(t: type[Any @ TagSet], partial: dict, model_field: Field):
    return "subparser"


@ovld
def make_argument(t: type[UnionAlias], partial: dict, model_field: Field):
    if any(issubclass(o, FieldModelizable) or TagSet.extract(o) for o in get_args(t)):
        return "subparser"
    else:
        options = [o for o in get_args(t) if o is not NoneType]
        if len(options) == 1:
            return recurse(options[0], partial, model_field)
        else:  # pragma: no cover
            raise TypeError("Unions of primitive/non-Modelizable types are not supported yet")


_add_argument_argnames = [
    "action",
    "nargs",
    "const",
    "default",
    "type",
    "choices",
    "required",
    "help",
    "metavar",
    "dest",
    "version",
    "alias",
    "option",
    "positional",
]


def add_argument_from_field(parser, fdest, overrides, field: Field):
    name = field.name.replace("_", "-")
    meta = {k: v for k, v in field.metadata.items() if k in _add_argument_argnames}
    overrides = dict(overrides)
    positional = meta.pop("positional", False) or overrides.pop("positional", False)
    fhelp = field.description or field.name
    mvar = name.split(".")[-1].upper()

    if positional:
        args = {"__args__": [fdest], "help": fhelp, "metavar": mvar, **overrides}
        if positional is not True:
            args["nargs"] = positional
        elif "nargs" not in args and "nargs" not in meta and "nargs" not in overrides:
            if not field.required:  # pragma: no cover
                args["nargs"] = "?"
    else:
        args = {
            "__args__": [f"--{name}" if len(name) > 1 else f"-{name}"],
            "help": fhelp,
            "metavar": mvar,
            "required": field.required,
            "dest": fdest,
            **meta,
            **overrides,
        }
    if "nargs" in args:
        try:
            args["nargs"] = int(args["nargs"])
        except (ValueError, TypeError):
            pass
    args = make_argument(field.type, args, field)
    if args == "subparser":
        add_arguments(field.type, parser, fdest, overrides.get("required", None) is False)
    else:
        pos = args.pop("__args__")
        if opt := args.pop("option", None):
            if not isinstance(opt, list):
                opt = [opt]
            pos[:] = opt
        if alias := args.pop("alias", None):
            if not isinstance(alias, list):
                alias = [alias]
            pos.extend([a for a in alias if a not in pos])
        parser.add_argument(*pos, **args)


@ovld
def add_arguments(
    t: type[FieldModelizable], parser: argparse.ArgumentParser, dest: str, partial: bool
):
    m = model(t)
    for fld in m.fields:
        if fld.name.startswith("_"):  # pragma: no cover
            continue
        fdest = _compose(dest, fld.name)
        overrides = {"required": False} if partial else {}
        add_argument_from_field(parser, fdest, overrides, fld)
    return parser


@ovld
def derive_options(t: type[Any @ TagSet]):
    base, ts = decompose(t)
    return list(ts.iterate(base))


@ovld
def derive_options(li: list):
    opts = []
    for o in li:
        opts.extend(recurse(o))
    return opts


@ovld
def add_arguments(t: type[UnionAlias], parser: argparse.ArgumentParser, dest: str, partial: bool):
    opts = [o for o in get_args(t) if o is not NoneType]
    if len(opts) == 1:
        recurse(opts[0], parser, dest, partial)
    else:
        recurse(derive_options(opts), parser, dest, partial)


@ovld(priority=1)
def add_arguments(
    t: type[Any @ TagSet], parser: argparse.ArgumentParser, dest: str, partial: bool
):
    recurse(derive_options(t), parser, dest, partial)


@ovld
def add_arguments(options: list, parser: argparse.ArgumentParser, dest: str, partial: bool):
    subparsers = parser.add_subparsers(dest=_compose(dest, tag_field), required=True)
    for tag, cls in options:
        subparser = subparsers.add_parser(tag, help=f"{strip(cls).__doc__ or tag}")
        recurse(cls, subparser, dest, partial)


@dataclass
class CLIDefinition:
    root_type: type = None
    mapping: dict[str, str | dict[str, Any]] = field(default_factory=lambda: {"": {"auto": True}})
    argparser: argparse.ArgumentParser = None
    description: str = None

    def __post_init__(self):
        if self.argparser is None:
            if self.description is None:
                description = (
                    strip(self.root_type).__doc__ or f"Arguments for {clsstring(self.root_type)}"
                )
            else:
                description = self.description
        self.argparser = argparse.ArgumentParser(description=description)
        for k, v in self.mapping.items():
            fld = field_at(self.root_type, k)
            if isinstance(v, str):
                v = {"__args__": [v]}
            elif v.pop("auto", False):
                add_arguments(fld.type, self.argparser, k, bool(v))
                if not v:
                    continue
            add_argument_from_field(self.argparser, k, v, fld)

    def __call__(self, argv):
        ns = self.argparser.parse_args(argv)
        values = {k: v for k, v in vars(ns).items() if v is not None}
        if (root := values.pop("", None)) is not None:
            vals = Sources(root, unflatten(values))
        else:
            vals = unflatten(values)
        return vals


def parse_cli(root_type, argv=None, mapping=None, description=None):
    mapping = {"": {"auto": True}} if mapping is None else mapping
    argv = sys.argv[1:] if argv is None else argv
    parser = CLIDefinition(root_type=root_type, mapping=mapping, description=description)
    return parser(argv)


class FromArguments(Medley):
    @ovld(priority=1)
    def deserialize(self, t: Any, obj: CommandLineArguments, ctx: Context):
        vals = obj.parse(t, obj.arguments)
        return recurse(t, vals, ctx)



from __future__ import annotations

import itertools
import json
from dataclasses import dataclass

import pytest
from apischema import serialize

from .interfaces import (
    adaptix,
    apischema,
    marshmallow,
    mashumaro,
    pydantic,
    serde,
    serieux,
)


@dataclass
class Point:
    x: int
    y: int


@dataclass
class Tree:
    left: Tree | int
    right: Tree | int


@dataclass
class Citizen:
    name: str
    birthyear: int
    hometown: str


@dataclass
class Country:
    languages: list[str]
    capital: str
    population: int
    citizens: list[Citizen]


@dataclass
class World:
    countries: dict[str, Country]


point = Point(x=17, y=83)


canada = Country(
    languages=["English", "French"],
    capital="Ottawa",
    population=39_000_000,
    citizens=[
        Citizen(
            name="Olivier",
            birthyear=1985,
            hometown="Montreal",
        ),
        Citizen(
            name="Abraham",
            birthyear=2018,
            hometown="Shawinigan",
        ),
    ],
)


world = World(countries={"canada": canada})


roboland = Country(
    languages=[f"Robolang{i}" for i in range(10000)],
    capital="Robopolis",
    population=1000,
    citizens=[
        Citizen(
            f"Bobot{i}",
            birthyear=3000 + i,
            hometown=f"Bobotown{i}",
        )
        for i in range(1000)
    ],
)


tree = Tree(1, Tree(Tree(2, 3), Tree(4, Tree(5, Tree(6, Tree(7, Tree(8, 9)))))))


id_to_thing = {id(v): k for k, v in globals().items()}


def bench(interfaces, data):
    cases = list(itertools.product(data, interfaces))
    return pytest.mark.parametrize(
        "data,interface",
        cases,
        ids=[f"{id_to_thing[id(d)]},{id_to_thing[id(i)]}" for d, i in cases],
    )


@bench(
    interfaces=[
        apischema,
        marshmallow,
        pydantic,
        adaptix,
        mashumaro,
        serieux,
        serde,
    ],
    data=[point, world, roboland, tree],
)
def test_serialize(interface, data, benchmark):
    fn = interface.serializer_for_type(type(data))
    result = benchmark(fn, data)
    assert result == serialize(type(data), data)


@bench(
    interfaces=[
        apischema,
        marshmallow,
        pydantic,
        adaptix,
        mashumaro,
        serieux,
        serde,
    ],
    data=[roboland],
)
def test_json(interface, data, benchmark):
    fn = interface.json_for_type(type(data))
    result = benchmark(fn, data)
    assert json.loads(result) == serialize(type(data), data)


@bench(
    interfaces=[
        apischema,
        marshmallow,
        pydantic,
        adaptix,
        mashumaro,
        serieux,
        serde,
    ],
    data=[point, world, roboland, tree],
)
def test_deserialize(interface, data, benchmark):
    data_ser = serialize(data)
    fn = interface.deserializer_for_type(type(data))
    result = benchmark(fn, data_ser)
    assert result == data



import json
from dataclasses import dataclass
from typing import Annotated, Any

import pytest

from serieux import Serieux, schema
from serieux.exc import ValidationError
from serieux.features.dotted import DottedNotation
from serieux.features.partial import Sources
from serieux.features.tagset import (
    FromEntryPoint,
    ReferencedClass,
    TagDict,
    TaggedSubclass,
    TagSetFeature,
    tag_field,
)

from ..definitions import Point

featured = (Serieux + TagSetFeature)()
serialize = featured.serialize
deserialize = featured.deserialize


@dataclass
class Animal:
    name: str


@dataclass
class Cat(Animal):
    selfishness: int

    def cry(self):
        return "me" * self.selfishness + "ow"


@dataclass
class HouseCat(Cat):
    cute: bool = True


@dataclass
class Wolf(Animal):
    size: int

    def cry(self):
        "a-woo" + "o" * self.size


@pytest.fixture
def chalou():
    td = TagDict()
    td.register("chat", Cat)
    td.register("loup", Wolf)
    return td


def test_tagdict_get_type(chalou):
    assert chalou.get_type("chat") is Cat
    assert chalou.get_type("loup") is Wolf


def test_tagdict_get_type_invalid(chalou):
    with pytest.raises(ValidationError):
        chalou.get_type("dog")


def test_tagdict_get_tag(chalou):
    assert chalou.get_tag(Cat) == "chat"
    assert chalou.get_tag(Wolf) == "loup"


def test_tagdict_get_tag_invalid(chalou):
    with pytest.raises(ValidationError):
        chalou.get_tag(HouseCat)


def test_tagdict_iterate(chalou):
    items = set(chalou.iterate(Animal))
    assert items == {("chat", Cat), ("loup", Wolf)}


def test_tagdict_register():
    tagset1 = TagDict()
    tagset1.register("chat", Cat)
    tagset1.register("loup", Wolf)
    assert tagset1.possibilities == {"chat": Cat, "loup": Wolf}

    tagset2 = TagDict()
    tagset2.register("chat")(Cat)
    tagset2.register("loup")(Wolf)
    assert tagset2.possibilities == {"chat": Cat, "loup": Wolf}

    tagset3 = TagDict()
    tagset3.register(Cat)
    tagset3.register(Wolf)
    assert tagset3.possibilities == {"cat": Cat, "wolf": Wolf}


def test_tagdict_serialize(chalou):
    cat = Cat(name="Kitty", selfishness=3)
    ser = serialize(Annotated[Animal, chalou], cat)
    assert ser == {
        tag_field: "chat",
        "name": "Kitty",
        "selfishness": 3,
    }


def test_tagdict_deserialize(chalou):
    ser = {
        tag_field: "chat",
        "name": "Kitty",
        "selfishness": 3,
    }
    cat = Cat(name="Kitty", selfishness=3)
    deser = deserialize(Annotated[Animal, chalou], ser)
    assert deser == cat


def test_tagdict_ser_deser2(chalou):
    wolf = Wolf(name="Wolfie", size=5)
    ser_wolf = serialize(Annotated[Animal, chalou], wolf)
    assert ser_wolf == {
        tag_field: "loup",
        "name": "Wolfie",
        "size": 5,
    }
    deser_wolf = deserialize(Annotated[Animal, chalou], ser_wolf)
    assert deser_wolf == wolf


def test_tagdict_default():
    td = TagDict({"default": Cat, "loup": Wolf})
    cat = Cat(name="Kitty", selfishness=3)
    ser_cat = serialize(Annotated[Animal, td], cat)
    assert ser_cat == {
        "name": "Kitty",
        "selfishness": 3,
    }
    deser_cat = deserialize(Annotated[Animal, td], ser_cat)
    assert deser_cat == cat


def test_tagdict_schema(file_regression, chalou):
    sch = schema(Annotated[Animal, chalou])
    file_regression.check(json.dumps(sch.compile(), indent=4))


def test_merge_tagsets(file_regression):
    cha = TagDict({"chat": Cat})
    lou = TagDict({"loup": Wolf})

    cat = Cat(name="Kitty", selfishness=3)
    wolf = Wolf(name="Wolfie", size=5)

    ser_cat = serialize(Annotated[Animal, cha, lou], cat)
    assert ser_cat == {
        tag_field: "chat",
        "name": "Kitty",
        "selfishness": 3,
    }
    deser_cat = deserialize(Annotated[Animal, cha, lou], ser_cat)
    assert deser_cat == cat

    ser_wolf = serialize(Annotated[Animal, cha, lou], wolf)
    assert ser_wolf == {
        tag_field: "loup",
        "name": "Wolfie",
        "size": 5,
    }
    deser_wolf = deserialize(Annotated[Animal, cha, lou], ser_wolf)
    assert deser_wolf == wolf

    sch = schema(Annotated[Animal, cha, lou])
    file_regression.check(json.dumps(sch.compile(), indent=4))

    housecat = HouseCat(name="Mimi", selfishness=8)
    with pytest.raises(ValidationError, match="No tagset could resolve for type"):
        serialize(Annotated[Animal, cha, lou], housecat)

    with pytest.raises(ValidationError, match="No tagset could resolve the tag"):
        deserialize(
            Annotated[Animal, cha, lou],
            {
                tag_field: "housecat",
                "name": "Mimi",
                "selfishness": 8,
            },
        )


def test_tagged_subclass():
    orig = Wolf(name="Wolfie", size=10)
    ser = serialize(TaggedSubclass[Animal], orig)
    assert ser == {
        tag_field: "tests.features.test_tagset:Wolf",
        "name": "Wolfie",
        "size": 10,
    }
    deser = deserialize(TaggedSubclass[Animal], ser)
    assert deser == orig


def test_serialize_not_top_level():
    @dataclass
    class Lynx:
        name: str
        selfishness: int

    orig = Lynx(name="Lina", selfishness=9)
    with pytest.raises(ValidationError, match="Only top-level symbols"):
        serialize(TaggedSubclass[Lynx], orig)


def test_serialize_wrong_class():
    orig = Wolf(name="Wolfie", size=10)
    with pytest.raises(ValidationError, match="Wolf.*is not a subclass of.*Cat"):
        serialize(TaggedSubclass[Cat], orig)


def test_deserialize_wrong_class():
    orig = {tag_field: "tests.features.test_tagset:Wolf", "name": "Wolfie", "size": 10}
    with pytest.raises(ValidationError, match="Wolf.*is not a subclass of.*Cat"):
        deserialize(TaggedSubclass[Cat], orig)


def test_resolve_default():
    ser = {"name": "Kevin"}
    assert deserialize(TaggedSubclass[Animal], ser) == Animal(name="Kevin")


def test_resolve_same_file():
    ser = {tag_field: "Cat", "name": "Katniss", "selfishness": 3}
    assert deserialize(TaggedSubclass[Animal], ser) == Cat(name="Katniss", selfishness=3)


def test_not_found():
    with pytest.raises(ValidationError, match="no attribute 'Bloop'"):
        ser = {tag_field: "Bloop", "name": "Quack"}
        deserialize(TaggedSubclass[Animal], ser)


def test_bad_resolve():
    with pytest.raises(ValidationError, match="Bad format for class reference"):
        ser = {tag_field: "x:y:z", "name": "Quack"}
        deserialize(TaggedSubclass[Animal], ser)


@dataclass
class Animals:
    alpha: TaggedSubclass[Animal]
    betas: list[TaggedSubclass[Animal]]


def test_tagged_subclass_partial():
    animals = deserialize(
        Animals,
        Sources(
            {
                "alpha": {
                    tag_field: "tests.features.test_tagset:Wolf",
                    "name": "Wolfie",
                    "size": 10,
                },
                "betas": [],
            },
        ),
    )
    assert isinstance(animals.alpha, Wolf)


def test_tagged_subclass_partial_merge():
    animals = deserialize(
        Animals,
        Sources(
            {
                "alpha": {
                    tag_field: "tests.features.test_tagset:Wolf",
                    "name": "Wolfie",
                    "size": 10,
                },
                "betas": [],
            },
            {"alpha": {tag_field: "tests.features.test_tagset:Wolf", "size": 13}},
        ),
    )
    assert isinstance(animals.alpha, Wolf)
    assert animals.alpha.name == "Wolfie"
    assert animals.alpha.size == 13


def test_tagged_subclass_partial_merge_subclass_left():
    animals = deserialize(
        Animals,
        Sources(
            {"alpha": {"name": "Roar"}},
            {
                "alpha": {
                    tag_field: "tests.features.test_tagset:Wolf",
                    "size": 10,
                },
                "betas": [],
            },
        ),
    )
    assert isinstance(animals.alpha, Wolf)
    assert animals.alpha.name == "Roar"
    assert animals.alpha.size == 10


def test_tagged_subclass_partial_merge_subclass_right():
    animals = deserialize(
        Animals,
        Sources(
            {
                "alpha": {
                    tag_field: "tests.features.test_tagset:Wolf",
                    "name": "Wolfie",
                    "size": 10,
                },
                "betas": [],
            },
            {"alpha": {"name": "Roar"}},
        ),
    )
    assert isinstance(animals.alpha, Wolf)
    assert animals.alpha.name == "Roar"
    assert animals.alpha.size == 10


def test_tagged_subclass_schema(file_regression):
    sch = schema(TaggedSubclass[Animal])
    file_regression.check(json.dumps(sch.compile(), indent=4))


def test_tagged_subclass_schema_fully_qualified(file_regression):
    sch = schema(Annotated[Animal, ReferencedClass])
    file_regression.check(json.dumps(sch.compile(), indent=4))


def test_open_schema():
    sch = schema(Annotated[Any, ReferencedClass])
    assert sch.compile(root=False) == {"type": "object", "additionalProperties": True}


def test_from_entry_points():
    # NOTE: serieux needs to be properly installed, e.g. with pip install or uv sync,
    # for the entry points to be registered
    OptF = Annotated[Any, FromEntryPoint("serieux.optional_features")]
    ser = serialize(OptF, DottedNotation())
    assert ser == {tag_field: "dotted"}
    assert deserialize(OptF, ser) == DottedNotation()
    sch = schema(OptF).compile(root=False)
    possibilities = {x["properties"][tag_field]["const"] for x in sch["oneOf"]}
    assert "dotted" in possibilities
    assert "autotag" in possibilities
    assert "include_file" in possibilities


def test_from_entry_points_with_default():
    # NOTE: serieux needs to be properly installed, e.g. with pip install or uv sync,
    # for the entry points to be registered
    OptF = Annotated[Any, FromEntryPoint("serieux.optional_features", default=DottedNotation)]
    ser = serialize(OptF, DottedNotation())
    assert ser == {}
    assert deserialize(OptF, {}) == DottedNotation()
    assert deserialize(OptF, {tag_field: "dotted"}) == DottedNotation()
    sch = schema(OptF).compile(root=False)
    assert any(
        ((tag_field not in x.get("properties", {})) or x.get("required") == [])
        for x in sch["oneOf"]
    )
    possibilities = {
        x["properties"][tag_field]["const"]
        for x in sch["oneOf"]
        if tag_field in x.get("properties", {})
    }
    assert "dotted" in possibilities
    assert "autotag" in possibilities
    assert "include_file" in possibilities


def test_from_entry_points_with_wrap():
    # NOTE: serieux needs to be properly installed, e.g. with pip install or uv sync,
    # for the entry points to be registered

    sub = {DottedNotation: Point}

    def wrap(x):
        return sub.get(x, x)

    OptF = Annotated[Any, FromEntryPoint("serieux.optional_features", wrap=wrap)]
    deser = deserialize(OptF, {tag_field: "dotted", "x": 1, "y": 2})
    assert deser == Point(1, 2)


def test_from_entry_points_errors():
    # NOTE: serieux needs to be properly installed, e.g. with pip install or uv sync,
    # for the entry points to be registered

    OptF = Annotated[Any, FromEntryPoint("serieux.optional_features")]

    with pytest.raises(ValidationError, match="No tag provided for entry point lookup"):
        deserialize(OptF, {})

    with pytest.raises(ValidationError, match="is not registered in entry point group"):
        deserialize(OptF, {tag_field: "not_a_real_entry_point"})

    fe = FromEntryPoint("serieux.optional_features")

    class NotRegistered:
        pass

    with pytest.raises(ValidationError, match="No entry point tag is registered for type"):
        fe.get_tag(NotRegistered, ctx=None)



from dataclasses import dataclass

from ovld import Medley
from ovld.dependent import Regexp
from rich.pretty import pprint

from serieux import Context, Serieux, deserialize, serialize

##################
# Implementation #
##################


@dataclass
class RGB:
    red: int
    green: int
    blue: int


@Serieux.extend
class RGBSerializer(Medley):
    """Custom serializer for RGB colors to/from hex strings."""

    def serialize(self, t: type[RGB], obj: RGB, ctx: Context):
        """Convert RGB to hex string format #rrggbb."""
        return f"#{obj.red:02x}{obj.green:02x}{obj.blue:02x}"

    def deserialize(self, t: type[RGB], obj: Regexp[r"^#[0-9a-fA-F]{6}$"], ctx: Context):
        """Convert hex string #rrggbb to RGB."""
        hex_str = obj.lstrip("#")
        red = int(hex_str[0:2], 16)
        green = int(hex_str[2:4], 16)
        blue = int(hex_str[4:6], 16)
        return RGB(red=red, green=green, blue=blue)


#################
# Demonstration #
#################


def main():
    color = RGB(red=255, green=128, blue=0)
    serialized = serialize(RGB, color)
    pprint({"original": color, "serialized": serialized}, expand_all=True)
    assert serialized == "#ff8000"

    scolor = "#ff8000"
    deserialized = deserialize(RGB, scolor)
    pprint({"serialized": scolor, "deserialized": deserialized}, expand_all=True)
    assert deserialized == color

    colors = [RGB(255, 0, 0), RGB(0, 255, 0), RGB(0, 0, 255)]
    serialized_list = serialize(list[RGB], colors)
    pprint({"colors": colors, "serialized": serialized_list})
    assert serialized_list == ["#ff0000", "#00ff00", "#0000ff"]

    scolors = ["#ff0000", "#00ff00", "#0000ff"]
    deserialized_list = deserialize(list[RGB], scolors)
    pprint({"serialized": scolors, "deserialized": deserialized_list})
    assert deserialized_list == colors


if __name__ == "__main__":
    main()



from typing import Annotated, Literal

from serieux import deserialize, schema, serialize
from serieux.instructions import Instruction, inherit, pushdown, strip

from .common import one_test_per_assert
from .features.test_usermeth import RGB

Apple = Instruction("Apple", annotation_priority=1)
Banana = Instruction("Banana", annotation_priority=2)
Carrot = Instruction("Carrot", annotation_priority=3)
Dog = Instruction("Dog", annotation_priority=4, inherit=False)
Useless = Instruction("Useless", annotation_priority=1, inherit=True)


@one_test_per_assert
def test_typetag_strip():
    assert strip(Apple[int], Apple) is int
    assert strip(Apple[Banana[int]], Apple) is Banana[int]
    assert strip(Banana[Apple[int]], Banana) is Apple[int]
    assert strip(Banana[int], Banana) is int
    assert strip(int, Apple) is int


@one_test_per_assert
def test_pushdown():
    assert pushdown(int) is int
    assert pushdown(Apple[int]) is int
    assert pushdown(Apple[list[int]]) == list[Apple[int]]
    assert pushdown(Apple[Banana[list[int]]]) == list[Apple[Banana[int]]]
    assert pushdown(Apple[int | str]) == Apple[int] | Apple[str]


def test_pushdown_literal():
    assert pushdown(Apple[Literal[1, 2]]) == Literal[1, 2]


@one_test_per_assert
def test_pushdown_no_inherit():
    # Dog is a tag that is not inherited when pushing down
    assert pushdown(Dog[list[int]]) == list[int]
    assert pushdown(Dog[Apple[list[int]]]) == list[Apple[int]]


def test_ser_deser_ignores_them():
    assert serialize(Useless[RGB], RGB(1, 2, 3)) == "#010203"
    assert deserialize(Useless[RGB], "#010203") == RGB(1, 2, 3)
    s1 = schema(Useless[RGB]).json()
    s2 = schema(RGB).json()
    assert s1 == s2


@one_test_per_assert
def test_inherit():
    assert inherit(int @ Banana @ Apple, float) == float @ Banana @ Apple
    assert inherit(int @ Apple @ Dog, float) == float @ Apple
    assert inherit(int @ Dog, float) is float
    assert inherit(int @ Banana @ Dog @ Apple, float) == float @ Banana @ Apple
    assert inherit(int, float) is float
    assert inherit(Annotated[int, "unrelated"], float) is float
    assert inherit(Annotated[int, Apple, "unrelated"], float) == float @ Apple
